import functools
import logging

import pandas as pd

# from elecsim.scen_error.scenario_data import wind_capacity_factor, solar_capacity_factor, historical_hourly_demand, \
#     segment_demand, load_duration_curve_diff, hydro_capacity_factor
import elecsim.scenario.scenario_data

logger = logging.getLogger(__name__)
"""
File name: solar_capacity_factor
Date created: 27/12/2018
Feature: # Calculates the average capacity factor per demand segment
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"


@functools.lru_cache(maxsize=512)
def get_capacity_factor(renewable_type, demand_hour):
        historical_demand = elecsim.scenario.scenario_data.historical_hourly_demand
        renewable_type = renewable_type.lower()
        capacity_data = get_capacity_data(renewable_type)
        if renewable_type in ['onshore', 'offshore', 'pv']:
            capacity_factor = segment_capacity_data_by_load_curve(capacity_data, historical_demand, renewable_type)
            capacity_factor_value = get_capacity_factor_value_for_segment(capacity_factor, demand_hour,
                                                                      renewable_type)
        else:
            capacity_factor_value = capacity_data
        return capacity_factor_value

def get_capacity_data(renewable_type):
    if renewable_type in ["onshore", 'offshore']:
        capacity_data = elecsim.scenario.scenario_data.wind_capacity_factor[['time', renewable_type]]
    elif renewable_type == "pv":
        capacity_data = elecsim.scenario.scenario_data.solar_capacity_factor
    elif renewable_type == 'hydro':
        capacity_data = elecsim.scenario.scenario_data.hydro_capacity_factor
    else:
        raise ValueError("Calculating demand factor can only be done for Onshore, Offshore or PV power generators.")
    return capacity_data

def segment_capacity_data_by_load_curve(capacity_data, historical_demand, renewable_type):
    demand_capacity = capacity_data.join(historical_demand, how='inner').dropna()
    demand_capacity = demand_capacity[elecsim.scenario.scenario_data.segment_demand[-1] < demand_capacity.demand]
    demand_capacity = demand_capacity[elecsim.scenario.scenario_data.segment_demand[0] > demand_capacity.demand]
    capacity_factor = demand_capacity.groupby(pd.cut(demand_capacity.demand, 20))[renewable_type].mean()
    capacity_factor = capacity_factor.to_frame()
    capacity_factor['diff'] = elecsim.scenario.scenario_data.load_duration_curve_diff['hours'].sort_values(ascending=False).values
    return capacity_factor

def get_capacity_factor_value_for_segment(capacity_factor, demand_hour, renewable_type):
    subsetted_dataframe = capacity_factor[capacity_factor['diff'] >= demand_hour].tail(1)
    capacity_factor_value = float(subsetted_dataframe[renewable_type].values)
    return capacity_factor_value










import logging
from itertools import chain
from random import shuffle

from mesa.time import BaseScheduler

logging.getLogger(__name__)

from elecsim.agents.generation_company.gen_co import GenCo
from elecsim.agents.demand.demand import Demand
from elecsim.market.electricity.power_exchange import PowerExchange

class OrderedActivation(BaseScheduler):
    """ A scheduler which activates each agent in the order that they are added to the scheduler

    Assumes that all agents have a step(model) method.

    """
    def step(self):
        """ Executes the step of all agents, one at a time.
        """

        gen_cos = [agent for agent in self.agents if isinstance(agent, GenCo)]
        shuffle(gen_cos)

        demand_agents = [agent for agent in self.agents if isinstance(agent, Demand)]
        power_exchange = [agent for agent in self.agents if isinstance(agent, PowerExchange)]

        logging.info("Stepping agents")
        for agent in chain(power_exchange, gen_cos, demand_agents):
            agent.step()
        self.steps += 1
        self.time += 1



import logging
from itertools import chain

import pandas as pd

from elecsim.market.electricity.bid import Bid
import elecsim.scenario.scenario_data

logger = logging.getLogger(__name__)



"""power_exchange.py: Functionality to run power exchange"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "Alexander@Kell.es"


class PowerExchange:

    def __init__(self, model):
        """
        Power exchange agent which contains functionality to tender and respond to bids.
        :param model: Model in which the agents are contained in.
        """
        self.model = model
        self.hold_duration_curve_prices = []
        self.price_duration_curve = pd.DataFrame(columns = ["year", "segment_hour", "segment_demand", "accepted_price"])

    def tender_bids(self, segment_hours, segment_demand, predict=False):
        """
        Function which iterates through the generator companies, requests their bids, orders them in order of price,
        and accepts bids.
        :param agents: All agents from simulation model.
        :param segment_hours: Value for number of hours particular electricity generation is required.
        :param segment_demand: Size of electricity consumption required.
        :param predict: Boolean that states whether the bids being tendered are for predicting price duration curve or whether it is for real bids.
        :return: None
        """
        agent = self.model.schedule.agents
        # generator_companies = [x for x in agent if isinstance(x, GenCo)]  # Selection of generation company agents
        generator_companies = [x for x in agent if hasattr(x, 'plants')]  # Selection of generation company agents

        # self.adjust_load_duration_curve_for_renewables()
        for segment_hour, segment_demand in zip(segment_hours, segment_demand):
            bids = []
            for generation_company in generator_companies:
                bids.append(generation_company.calculate_bids(segment_hour, predict))
            sorted_bids = self._sort_bids(bids)
            accepted_bids = self._respond_to_bids(sorted_bids, segment_hour, segment_demand)

            logger.debug("segment hour: {}".format(segment_hour))
            self._accept_bids(accepted_bids)
            highest_bid = max(bid.price_per_mwh for bid in accepted_bids)

            self._create_load_duration_price_curve(segment_hour, segment_demand, highest_bid)

        self.price_duration_curve = pd.DataFrame(self.hold_duration_curve_prices)
        if predict:
            logger.debug("predicted self.price_duration_curve: {}".format(self.price_duration_curve))
        else:
            logger.info("actual self.price_duration_curve: {}".format(self.price_duration_curve))

        return self.price_duration_curve[self.price_duration_curve.year == self.model.year_number].accepted_price.mean()

    def _create_load_duration_price_curve(self, segment_hour, segment_demand, accepted_price):
        segment_price_data = {
                'year': self.model.year_number,
                'segment_hour': segment_hour,
                'segment_demand': segment_demand,
                'accepted_price': accepted_price
            }

        self.hold_duration_curve_prices.append(segment_price_data)

    @staticmethod
    def _accept_bids(accepted_bids):
        # highest_accepted_bid = accepted_bids[-1].price_per_mwh
        highest_accepted_bid = max(bid.price_per_mwh for bid in accepted_bids)
        for bids in accepted_bids:

            # logger.debug("bid price: {}, plant name: {}, plant capacity: {}".format(bids.price_per_mwh, bids.plant.name, bids.plant.capacity_mw))
            bids.price_per_mwh = highest_accepted_bid

    @staticmethod
    def _sort_bids(bids):
        """
        Sorts bids in order of price
        :param bids: Bid objects
        :return: Return bids in order of price
        """
        bids = list(chain.from_iterable(bids))
        sorted_bids = sorted(bids, key=lambda x: x.price_per_mwh)
        return sorted_bids

    def _respond_to_bids(self, bids, segment_hour, capacity_required):
        """
        Response to bids based upon price and capacity required. Accepts bids in order of cheapest generator.
        Continues to accept bids until capacity is met for those hours.
        :param bids: Bid objects.
        :param capacity_required: Capacity required for this segment.
        :return: Returns a list of bids which have been accepted.
        """
        accepted_bids = []

        for bid in bids:
            # logger.debug('bid: price: {}'.format(bid.price_per_mwh))
            if capacity_required > 0 and capacity_required > bid.capacity_bid:
                bid.accept_bid(segment_hour)
                capacity_required -= bid.capacity_bid
                accepted_bids.append(bid)
                bid.plant.accepted_bids.append(bid)
                logger.debug('bid ACCEPTED: price: {}, year: {}, capacity required: {}, capacity: {}, capacity_bid: {}, type: {}, name {}'.format(bid.price_per_mwh, bid.plant.construction_year, capacity_required, bid.plant.capacity_mw, bid.capacity_bid, bid.plant.plant_type,  bid.plant.name))
            elif bid.capacity_bid > capacity_required > 0:
                bid.partially_accept_bid(segment_hour, capacity_required)
                capacity_required = 0
                accepted_bids.append(bid)
                bid.plant.accepted_bids.append(bid)
                logger.debug('bid PARTIALLY ACCEPTED: price: {}, year: {}, capacity required: {}, capacity: {}, capacity_bid: {}, type: {}, name {}'.format(bid.price_per_mwh, bid.plant.construction_year, capacity_required, bid.plant.capacity_mw, bid.capacity_bid, bid.plant.plant_type,  bid.plant.name))
            else:
                bid.reject_bid(segment_hour=segment_hour)
                logger.debug('bid REJECTED: price: {}, year: {}, capacity required: {}, capacity: {}, capacity_bid: {}, type: {}, name {}'.format(bid.price_per_mwh, bid.plant.construction_year, capacity_required, bid.plant.capacity_mw, bid.capacity_bid, bid.plant.plant_type,  bid.plant.name))
        if capacity_required > 0:
            accepted_bids.append(Bid(None, None, segment_hour, 0, elecsim.scenario.scenario_data.lost_load, self.model.year_number))
        return accepted_bids




import logging
from functools import lru_cache

import numpy as np

from elecsim.market.electricity.power_exchange import PowerExchange
from elecsim.role.market.latest_market_data import LatestMarketData
import elecsim.scenario.scenario_data

logger = logging.getLogger(__name__)

"""
File name: predict_load_duration_prices
Date created: 11/01/2019
Feature: # Predict Load Duration Prices
"""

__author__ = "Alexander Kell"
__copyright__ = "Copyright 2018, Alexander Kell"
__license__ = "MIT"
__email__ = "alexander@kell.es"



class PredictPriceDurationCurve:

    def __init__(self, model):
        self.model = model

    def predict_price_duration_curve(self, look_back_period):
        demand_change_predicted = LatestMarketData(self.model).agent_forecast_value("demand", look_back_period, elecsim.scenario.scenario_data.years_for_agents_to_predict_forward)
        predicted_consumption = [cons * demand_change_predicted for cons in self.model.demand.segment_consumption]

        power_ex = PowerExchange(self.model)
        power_ex.tender_bids(self.model.demand.segment_hours, predicted_consumption, predict=True)
        predicted_price_duration_curve = power_ex.price_duration_curve

        predicted_price_duration_curve = estimate_lost_load_price(predicted_price_duration_curve)

        logger.debug("predicted_price_duration_curve: \n {}".format(predicted_price_duration_curve))
        return predicted_price_duration_curve


def estimate_lost_load_price(predicted_price_duration_curve):
    if all(predicted_price_duration_curve.accepted_price==elecsim.scenario.scenario_data.lost_load):
        return predicted_price_duration_curve
    if any(predicted_price_duration_curve.accepted_price==elecsim.scenario.scenario_data.lost_load):
        predicted_price_duration_curve.accepted_price = predicted_price_duration_curve.accepted_price.replace(elecsim.scenario.scenario_data.lost_load,np.nan)
        # predicted_price_duration_curve.accepted_price.interpolate(method="polynomial", order=1)
        if predicted_price_duration_curve.accepted_price.count() > 1:
            predicted_price_duration_curve_training = predicted_price_duration_curve.dropna()

            p = np.poly1d(np.polyfit(predicted_price_duration_curve_training.segment_demand, predicted_price_duration_curve_training.accepted_price,1))
            extrapolated = p(predicted_price_duration_curve.loc[np.isnan(predicted_price_duration_curve.accepted_price), 'segment_demand'])
            logger.debug("extrapolated: {}".format(extrapolated))
            predicted_price_duration_curve.loc[np.isnan(predicted_price_duration_curve.accepted_price), "accepted_price"] = extrapolated
            return predicted_price_duration_curve
        elif predicted_price_duration_curve.accepted_price.count() == 1:
            predicted_price_duration_curve.accepted_price = predicted_price_duration_curve.accepted_price.fillna(predicted_price_duration_curve.accepted_price.mean())
            return predicted_price_duration_curve
    else:
        return predicted_price_duration_curve
@lru_cache(1024)
def get_price_duration_curve(model, look_back_period):
    predicted_price_duration_curve = PredictPriceDurationCurve(model).predict_price_duration_curve(look_back_period=look_back_period)
    return predicted_price_duration_curve

